/*
** Module   : SocketListenerImpl.java
** Abstract : A class for implementing the SERVER-SOCKET method and attributes
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ----------------------------------Description---------------------------------
** 001 CS  20130204 Created initial version.
** 002 CA  20130221 Added Deletable support. setType was renamed to setResourceType.
** 003 OM  20130304 Refactored isValid and isUnknown of WrappedResource to valid and unknown.
** 004 CS  20130404 Implement runtime support
*/

package com.goldencode.p2j.util;

import java.io.*;
import java.net.*;
import java.security.*;
import java.security.cert.*;
import javax.net.ssl.*;

/**
 * A class for implementing the SERVER-SOCKET method and attributes.
 */ 
public class SocketListenerImpl 
extends SensitiveResource
implements SocketListener,
           Sensitive
{
   private int maxConn = 0;
   
   /** Timeout for connection*/
   private int timeout = 180;
   
   /** default keyalias*/
   private String keyalias = "default_server";
   
   /** Boolean flag to keep listening*/
   private boolean keepListening = true;
   
   /** The server socket context*/
   private ServerSocket socketListener;
   
   /** The response procedure name*/
   private String connectProcedure;
   
   /** The response procedure context*/
   private handle contextHandle;
   
   /**
    * Default constructor which initializes SERVER-SOCKET specific settings.
    */
   public SocketListenerImpl()
   {
      // TODO implement constructor logic EX:
      setResourceType("SERVER-SOCKET");
      setUnknown(false);
   }
   
   /**
    * Constructor with PREV-SIBLING and NEXT-SIBLING as attributes.
    * 
    * @param prevSibling
    *        The previously created Socket handle
    * @param nextSibling
    *        The following Socket handle
    */
   public SocketListenerImpl(handle prevSibling, handle nextSibling)
   {
      this();
      setPrevSibling(prevSibling);
      setNextSibling(nextSibling);
   }

   /**
    * Reports if this object is valid for use.  
    *
    * @return   <code>true</code> if we are valid (can be used).
    */
   public boolean valid()
   {
      return true;
   }
   
   /**
    * Java implementation for the ENABLE-CONNECTION()  SERVER-SOCKET method, 
    * this will attempt to start listening  on the port specified by the 
    * option argument, and will return <code>true</code> if operation 
    * succeeded, or <code>false</code> otherwise.
    * <p>
    * The following options can be specified as valid ENABLE-CONNECTIONS() 
    * options:
    * <ul>
    *    <li><b>-S port</b> The port to which to connect.
    *    <li><b>-ssl</b> If specified the connection will be made using ssl.
    *    <li><b>-pf config-file-name</b> Specifies the file from which options
    *        will be taken.
    *    <li><b>-qsize maxconn</b> The maximum number of connections to handle 
    *        before refusing extra incoming connections.
    *    <li><b>-keyalias alias</b> The alias for the key and digital 
    *        certificates, default is : default_server.
    *   
    *    <li><b>-keyaliaspasswd pwd</b> The password for the key and digital
    *        certificate. To use always with -keyalias option.
    *    <li><b>-nossesioncache</b> Disables SSL client session cache.
    *    <li><b>-sessiontimeout time</b> The number of seconds before the SSL
    *        client request is timeout, default value is 180..
    * </ul>
    * 
    * @param  option
    *         The set of options that define the enable-connections method.
    * 
    * @return <code>true</code> if operation succeeded, or <code>false</code>
    *         otherwise
    */
   public logical enableConnections(String option)
   {
      return enableConnections(new character(option));
   }
   
   /**
    * Java implementation for the ENABLE-CONNECTION()  SERVER-SOCKET method, 
    * this will attempt to start listening  on the port specified by the 
    * option argument, and will return <code>true</code> if operation 
    * succeeded, or <code>false</code> otherwise.
    * <p>
    * The following options can be specified as valid ENABLE-CONNECTIONS() 
    * options:
    * <ul>
    *    <li><b>-S port</b> The port to which to connect.
    *    <li><b>-ssl</b> If specified the connection will be made using ssl.
    *    <li><b>-pf config-file-name</b> Specifies the file from which options
    *        will be taken.
    *    <li><b>-qsize maxconn</b> The maximum number of connections to handle 
    *        before refusing extra incoming connections.
    *    <li><b>-keyalias alias</b> The alias for the key and digital 
    *        certificates, default is : default_server.
    *   
    *    <li><b>-keyaliaspasswd pwd</b> The password for the key and digital
    *        certificate. To use always with -keyalias option.
    *    <li><b>-nossesioncache</b> Disables SSL client session cache.
    *    <li><b>-sessiontimeout time</b> The number of seconds before the SSL
    *        client request is timeout, default value is 180.
    * </ul>
    * 
    * @param  option
    *         The set of options that define the enable-connections method.
    * 
    * @return <code>true</code> if operation succeeded, or <code>false</code>
    *         otherwise.
    */
   public logical enableConnections(character option)
   {
      integer errNo = new integer();
      integer port = new integer();
      character keyAlias = new character();
      character fileOpts = new character(option);
      character errMsg = new character();
      character keyAliasPwd = new character();
      logical ssl = new logical(false);
      logical noSessionCache = new logical(false);
      
      getParams(fileOpts, errMsg, errNo, port, keyAlias, 
                keyAliasPwd, ssl, noSessionCache);
      
      if (!errMsg.isUnknown())
      {
         ErrorManager.recordOrThrowError(errNo.intValue(), errMsg.toStringMessage());
         ErrorManager.recordOrThrowError(5509, "Unable to process parameters.");
         ErrorManager.recordOrThrowError(5510, "Invalid parameter string.");
         return new logical(false);
      }
      
      if(!ssl.booleanValue())
      {
         try
         {
            socketListener = new ServerSocket(port.intValue());
            socketListener.setSoTimeout(timeout);
            socketListener.setReceiveBufferSize(maxConn);
            
            startListening();
         }
         catch (IOException e)
         {
            // TODO Auto-generated catch block
            e.printStackTrace();
         }
      }
      else
      {
         
         try
         {
            char[] keyPassword = null;
            FileInputStream keyFile = null;
            
            if (!keyAlias.isUnknown())
            {
               keyFile = new FileInputStream(keyAlias.toStringMessage());
            }
            else
            {
               keyFile = new FileInputStream(keyalias);
            }
            
            if (!keyAliasPwd.isUnknown())
            {
               keyPassword = keyAliasPwd.toStringMessage().toCharArray();
            }
            // init keystore
            KeyStore keyStore =  KeyStore.getInstance(KeyStore.getDefaultType());
            keyStore.load(keyFile, keyPassword);
          
            // init KeyManagerFactory
            KeyManagerFactory keyManagerFactory = null;
            
            keyManagerFactory = KeyManagerFactory.getInstance(KeyManagerFactory.getDefaultAlgorithm());
            keyManagerFactory.init(keyStore, keyPassword);

            // init KeyManager
            KeyManager keyManagers[] = keyManagerFactory.getKeyManagers();
            
            // init the SSL context
            SSLContext sslContext = null;
            sslContext = SSLContext.getDefault();

            sslContext.init(keyManagers, null, new SecureRandom());

            SSLServerSocketFactory factory = (SSLServerSocketFactory) SSLServerSocketFactory.getDefault();
            SSLServerSocket sslSocketListener = (SSLServerSocket) factory.createServerSocket(port.intValue(), maxConn);
            sslSocketListener.setSoTimeout(timeout);
            socketListener = sslSocketListener;

         }

         catch (KeyManagementException e1)
         {
            // TODO Auto-generated catch block
            e1.printStackTrace();
         }
         catch (NoSuchAlgorithmException e1)
         {
            // TODO Auto-generated catch block
            e1.printStackTrace();
         }
         catch (UnrecoverableKeyException e1)
         {
            // TODO Auto-generated catch block
            e1.printStackTrace();
         }
         catch (KeyStoreException e1)
         {
            // TODO Auto-generated catch block
            e1.printStackTrace();
         }
         catch (FileNotFoundException e)
         {
            // TODO Auto-generated catch block
            e.printStackTrace();
         }
         catch (IOException e)
         {
            // TODO Auto-generated catch block
            e.printStackTrace();
         }
         catch (CertificateException e)
         {
            // TODO Auto-generated catch block
            e.printStackTrace();
         }
      }
      
      return new logical(true);
   }

   /**
    * Java implementation for the DISABLE-CONNECTIONS() SERVER-SOCKET method, 
    * this will stop listening and accepting any new connections, but without
    * interrupting the existing ones that haven't finished, and returns 
    * <code>true</code> if succeeded or  <code>false</code> otherwise.
    * 
    * @return <code>true</code> if the operation succeeded or <code>false</code> 
    *         otherwise.
    */
   public logical disableConnections()
   {
      return new logical(true);
   }
   
   /**
    * Java implementation for the SET-CONNECT-PROCEDURE() SERVER-SOCKET method, 
    * this will set the procedure to execute when a CONNECT event is 
    * received.
    * <p>
    * The set-connect-procedure can also contain a handle to a higher 
    * level procedure which contains the read response procedure (to resolve 
    * any procedure context issues).
    * 
    * @param  procedureName
    *         The name of the procedure we want to set for read response.
    *         
    * @return <code>true</code> if the operation was successful or 
    *         <code>false</code> otherwise.
    */
   public logical setConnectProcedure(String procedureName)
   {
      return setConnectProcedure(new character(procedureName));
   }

   /**
    * Java implementation for the SET-CONNECT-PROCEDURE() SERVER-SOCKET method, 
    * this will set the procedure to execute when a CONNECT event is 
    * received.
    * <p>
    * The set-connect-procedure can also contain a handle to a higher 
    * level procedure which contains the read response procedure (to resolve 
    * any procedure context issues).
    * 
    * @param  procedureName
    *         The name of the procedure we want to set for read response.
    *         
    * @return <code>true</code> if the operation was successful or 
    *         <code>false</code> otherwise.
    */
   public logical setConnectProcedure(character procedureName)
   {
      return setConnectProcedure(procedureName, null);
   }

   /**
    * Java implementation for the SET-CONNECT-PROCEDURE() SERVER-SOCKET method, 
    * this will set the procedure to execute when a CONNECT event is 
    * received.
    * <p>
    * The set-connect-procedure can also contain a handle to a higher 
    * level procedure which contains the read response procedure (to resolve 
    * any procedure context issues).
    * 
    * @param  procedureName
    *         The name of the procedure we want to set for read response.
    * @param  procHandle
    *         The handle for the procedure containing the procedure we want
    *         to set as connect procedure - optional.
    *         
    * @return <code>true</code> if the operation was successful or 
    *         <code>false</code> otherwise.
    */
   public logical setConnectProcedure(String procedureName, handle procHandle)
   {
      return setConnectProcedure(new character(procedureName), procHandle);
   }

   /**
    * Java implementation for the SET-CONNECT-PROCEDURE() SERVER-SOCKET method, 
    * this will set the procedure to execute when a CONNECT event is 
    * received.
    * <p>
    * The set-connect-procedure can also contain a handle to a higher 
    * level procedure which contains the read response procedure (to resolve 
    * any procedure context issues).
    * 
    * @param  procedureName
    *         The name of the procedure we want to set for read response.
    * @param  procHandle
    *         The handle for the procedure containing the procedure we want
    *         to set as connect procedure - optional.
    *         
    * @return <code>true</code> if the operation was successful or 
    *         <code>false</code> otherwise.
    */
   public logical setConnectProcedure(character procedureName,
                                      handle procHandle)
   {
      if (!procedureName.isUnknown())
      {
         connectProcedure = procedureName.toStringMessage();
      }
      contextHandle = procHandle;
      return new logical(true);
   }

   /**
    * Delete the resource.
    */
   @Override
   protected void resourceDelete()
   {
      // TODO Auto-generated method stub
   }
   /**
    * Start receiving connections until told to end
    */
   private void startListening()
   {
      while (keepListening)
      {
            try
            {
               java.net.Socket socket = socketListener.accept();
               Socket s = new SocketImpl(socket);
               ControlFlowOps.invokeIn(connectProcedure, contextHandle, new handle(s));
            }
            catch (IOException e)
            {
               // TODO Auto-generated catch block
               e.printStackTrace();
            }
      }
      
   }

   /**
    * Utility method for parsing the enableConnections parameters
    * @param options
    *        The string to parse.
    * @param errMsg
    *        The error message to set.
    * @param errNo
    *        The error number to set.
    * @param port
    *        The port which will be parse.
    * @param keyAlias
    *        The alias for the key and digital certificates to be parsed.
    * @param keyAliasPwd
    *        The password for the key and digital certificate to parse.
    * @param ssl
    *        True if -ssl parameter will be parsed.
    * @param noSessionCache
    *        The -nosessioncache parameter that will be parsed.
    */
   private void getParams(character options,
                          character errMsg,
                          integer errNo,
                          integer port,
                          character keyAlias,
                          character keyAliasPwd,
                          logical ssl,
                          logical noSessionCache)
   {
    
      String strParams = options.toStringMessage();
      
      if (!strParams.toUpperCase().contains("-S") && !strParams.toUpperCase().contains("-PF"))
      {
         errMsg.assign("The -S parameter must be specified to connect");
         errNo.assign(5488);
      }
      
      String[] tokens = strParams.split(" ");
      
      for (int i = 0; i < tokens.length; i++)
      {
         if (!tokens[i].startsWith("-"))
         {
            errMsg.assign("Could not recognize arguments: " + tokens[i]);
            errNo.assign(301);
         }
         else
         {
            // if the option is the last element error
            if (i + 1 > tokens.length)
            {
               // Progress 4GL doesn't check if the argument is a valid accepted argument
               // this is just replicating the behavior
               errMsg.assign("You have not supplied a parameter for argument: " + tokens[i]);
               errNo .assign(1403);
               break;
            }
            if (tokens[i].equalsIgnoreCase("-S"))
            {
               try
               {
                  port.assign(Integer.parseInt(tokens[i+1]));
                  i++;
               }
               catch(NumberFormatException e)
               {
                  errMsg.assign("Unknown service " + tokens[i+1] + " transport TCP");
                  errNo.assign(5484);
                  break;
               }               
            }
            else if (tokens[i].equalsIgnoreCase("-ssl"))
            {
               ssl.assign(true);
            }
            else if (tokens[i].equalsIgnoreCase("-qsize"))
            {
               maxConn = Integer.parseInt(tokens[i+1]);
               i++;
            }
            else if (tokens[i].equalsIgnoreCase("-keyalias"))
            {
               keyAlias.assign(tokens[i+1]);
               i++;
            }
            else if (tokens[i].equalsIgnoreCase("-keyaliaspasswd"))
            {
               keyAliasPwd.assign(tokens[i+1]);
               i++;
            }
            else if (tokens[i].equalsIgnoreCase("-nossesioncache "))
            {
               noSessionCache.assign(true);
            }
            else if (tokens[i].equalsIgnoreCase("-sessiontimeout"))
            {
               timeout = Integer.parseInt(tokens[i+1]);
               i++;
            }
            else if (tokens[i].equalsIgnoreCase("-pf"))
            {
               File configFile = new File(tokens[i+1]);
               if (!configFile.exists() && configFile.isFile())
               {
                  errMsg.assign("Unable to open parameter file " + tokens[i+1] + ", errno 2.");
                  errNo.assign(1247);
                  break;
               }
               else
               {
                  // clear any existing errors
                  errMsg = new character();
                  BufferedReader inputReader = null;
                  try
                  {
                     inputReader = new BufferedReader(new FileReader(configFile));
                  }
                  catch (FileNotFoundException e)
                  {
                    
                  }
                  try
                  {
                     StringBuffer b =  new StringBuffer();
                     int asc = inputReader.read();
                     while (asc != -1)
                     {
                        char chr = (char) asc;
                        b.append(chr);
                        asc = inputReader.read();
                     }
                     character fileOpts = new character(b.toString());
                     getParams(fileOpts, errMsg, errNo, port, keyAlias, 
                               keyAliasPwd, ssl, noSessionCache);
                  }
                  catch (IOException e)
                  {
                     errMsg.assign("An error has occured while reading from the parameter file " +
                                   tokens[i+1] + ", errno 2.");
                  }
                  
               }
            }
         }
      }
   }
   
}
